# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('notify', '0004_auto_20170525_0233'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notification',
            name='actor_text',
            field=models.CharField(max_length=200, null=True, verbose_name='Anonymous text for actor', blank=True),
        ),
        migrations.AlterField(
            model_name='notification',
            name='obj_text',
            field=models.CharField(max_length=200, null=True, verbose_name='Anonymous text for action object', blank=True),
        ),
        migrations.AlterField(
            model_name='notification',
            name='target_text',
            field=models.CharField(max_length=200, null=True, verbose_name='Anonymous text for target', blank=True),
        ),
        migrations.AlterField(
            model_name='notification',
            name='verb',
            field=models.CharField(max_length=200, verbose_name='Verb of the action'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-05-25 02:33
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('notify', '0003_auto_20160325_1920'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notification',
            name='actor_text',
            field=models.CharField(blank=True, max_length=100, null=True, verbose_name='Anonymous text for actor'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='nf_type',
            field=models.CharField(default='default', max_length=40, verbose_name='Type of notification'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='obj_text',
            field=models.CharField(blank=True, max_length=100, null=True, verbose_name='Anonymous text for action object'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='target_text',
            field=models.CharField(blank=True, max_length=100, null=True, verbose_name='Anonymous text for target'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='verb',
            field=models.CharField(max_length=100, verbose_name='Verb of the action'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2015-12-21 07:37
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('notify', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notification',
            name='actor_content_type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='notify_actor', to='contenttypes.ContentType', verbose_name='Content type of actor object'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='actor_object_id',
            field=models.PositiveIntegerField(blank=True, null=True, verbose_name='ID of the actor object'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='actor_text',
            field=models.CharField(blank=True, max_length=50, null=True, verbose_name='Anonymous text for actor'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='actor_url_text',
            field=models.URLField(blank=True, null=True, verbose_name='Anonymous URL for actor'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='deleted',
            field=models.BooleanField(default=False, verbose_name='Soft delete status'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='description',
            field=models.CharField(blank=True, max_length=255, null=True, verbose_name='Description of the notification'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='extra',
            field=jsonfield.fields.JSONField(blank=True, null=True, verbose_name='JSONField to store addtional data'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='nf_type',
            field=models.CharField(default='default', max_length=20, verbose_name='Type of notification'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='obj_content_type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='notify_object', to='contenttypes.ContentType', verbose_name='Content type of action object'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='obj_object_id',
            field=models.PositiveIntegerField(blank=True, null=True, verbose_name='ID of the target object'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='obj_text',
            field=models.CharField(blank=True, max_length=50, null=True, verbose_name='Anonymous text for action object'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='obj_url_text',
            field=models.URLField(blank=True, null=True, verbose_name='Anonymous URL for action object'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='read',
            field=models.BooleanField(default=False, verbose_name='Read status'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='recipient',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='notifications', to=settings.AUTH_USER_MODEL, verbose_name='Notification receiver'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='target_content_type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='notify_target', to='contenttypes.ContentType', verbose_name='Content type of target object'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='target_object_id',
            field=models.PositiveIntegerField(blank=True, null=True, verbose_name='ID of the target object'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='target_text',
            field=models.CharField(blank=True, max_length=50, null=True, verbose_name='Anonymous text for target'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='target_url_text',
            field=models.URLField(blank=True, null=True, verbose_name='Anonymous URL for target'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='verb',
            field=models.CharField(max_length=50, verbose_name='Verb of the action'),
        ),
    ]



import os
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

SECRET_KEY = 'fake-key'

ROOT_URLCONF = 'notify.tests.urls'

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.sessions',
    'django.contrib.contenttypes',
    'django.contrib.staticfiles',
    'notify',
)

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
        'TEST_NAME': os.path.join(os.path.dirname(__file__), 'test.db'),
    }
}

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
)

AUTHENTICATION_BACKENDS = ('django.contrib.auth.backends.ModelBackend',)

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [os.path.join(BASE_DIR, 'templates')],

        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

STATIC_ROOT = os.path.abspath(os.path.join(BASE_DIR, 'static'))
STATIC_URL = '/static/'
STATICFILES_DIRS = (os.path.join(STATIC_ROOT, 'sitefiles'),)

MEDIA_ROOT = os.path.join(BASE_DIR, 'media')
MEDIA_URL = '/media/'



from django.conf.urls import url
from notify import views as nf

urlpatterns = [
    url(r'^all/$', nf.notifications, name="all"),
    url(r'^api/all/$', nf.NotificationsList.as_view(), name="api_list"),
    url(r'^api/update/$', nf.notification_update, name="update"),
    url(r'^mark/$', nf.mark, name='mark'),
    url(r'^mark-read/$', nf.mark_read, name='mark_read'),
    url(r'^mark-unread/$', nf.mark_unread, name='mark_unread'),
    url(r'^mark-all/$', nf.mark_all, name='mark_all'),
    url(r'^delete/$', nf.delete, name='delete'),
    url(r'^rdr/(?P<notification_id>[\d]+)/$', nf.read_and_redirect,
        name='read_and_redirect'),
]



from django.contrib import admin
from .models import Notification


@admin.register(Notification)
class NotificationAdmin(admin.ModelAdmin):
    list_display = ('recipient', 'actor', 'verb',
                    'obj', 'target', 'read', 'deleted')
    list_filter = ('read', 'deleted', 'nf_type', 'created')
    search_fields = ('recipient__username', 'verb', 'description')
    raw_id_fields = ('recipient', )

    fieldsets = (
            ('Basic details',
                {'fields': ('recipient', 'verb', 'description', 'nf_type')}),

            ('Actor details',
                {'fields': ('actor_content_type', 'actor_object_id',
                            'actor_text', 'actor_url_text')}),

            ('Object details',
                {'fields': ('obj_content_type', 'obj_object_id',
                            'obj_text', 'obj_url_text')}),

            ('Target details',
                {'fields': ('target_content_type', 'target_object_id',
                            'target_text', 'target_url_text')}),

            ('Other details',
                {'fields': ('extra', 'read', 'deleted')}),
        )



from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.conf import settings
from django.db.models import QuerySet
from jsonfield.fields import JSONField
from django.utils.encoding import python_2_unicode_compatible
from django.utils.html import escape
from django.utils.timesince import timesince
from django.utils.translation import ugettext_lazy as _
from django.utils.encoding import force_text
from django.utils.functional import cached_property

from .utils import prefetch_relations


class Response(object):
    HTTP_200_OK = 200
    HTTP_202_ACCEPTED = 202
    HTTP_400_BAD_REQUEST = 400
    HTTP_500_INTERNAL_SERVER_ERROR = 500


class NotificationQueryset(QuerySet):

    """
    Chain-able QuerySets using ```.as_manager()``.
    """

    def prefetch(self):
        """
        Marks the current queryset to prefetch all generic relations.
        """
        qs = self.select_related()
        qs._prefetch_relations = True
        return qs

    def _fetch_all(self):
        if self._result_cache is None:
            if hasattr(self, '_prefetch_relations'):
                # removes the flag since prefetch_relations is recursive
                del self._prefetch_relations
                prefetch_relations(self)
                self._prefetch_relations = True
        return super(NotificationQueryset, self)._fetch_all()

    def _clone(self, **kwargs):
        clone = super(NotificationQueryset, self)._clone(**kwargs)
        if hasattr(self, '_prefetch_relations'):
            clone._prefetch_relations = True
        return clone

    def active(self):
        """
        QuerySet filter() for retrieving both read and unread notifications
        which are not soft-deleted.

        :return: Non soft-deleted notifications.
        """
        return self.filter(deleted=False)

    def read(self):
        """
        QuerySet filter() for retrieving read notifications.

        :return: Read and active Notifications filter().
        """
        return self.filter(deleted=False, read=True)

    def unread(self):
        """
        QuerySet filter() for retrieving unread notifications.

        :return: Unread and active Notifications filter().
        """
        return self.filter(deleted=False, read=False)

    def unread_all(self, user=None):
        """
        Marks all notifications as unread for a user (if supplied)

        :param user: Notification recipient.

        :return: List of updated notifications with status codes in form [{"status": 200, "id": "5"}, ..]
        """
        response = []
        qs = self.read()
        if user:
            qs.filter(recipient=user)
        for notification in qs:
            if notification.read == False:
                response.append({'id': notification.id, 'status': Response.HTTP_202_ACCEPTED})
            else:
                notification.read = False
                notification.save()
                response.append({'id': notification.id, 'status': Response.HTTP_200_OK})
        return response

    def read_all(self, user=None):
        """
        Marks all notifications as read for a user (if supplied)

        :param user: Notification recipient.

        :return: List of updated notifications with status codes
        """
        response = []
        qs = self.unread()
        if user:
            qs.filter(recipient=user)
        for notification in qs:
            if notification.read == True:
                response.append({'id': notification.id, 'status': Response.HTTP_202_ACCEPTED})
            else:
                notification.read = True
                notification.save()
                response.append({'id': notification.id, 'status': Response.HTTP_200_OK})
        return response

    def delete_all(self, user=None):
        """
        Method to soft-delete all notifications of a User (if supplied)

        :param user: Notification recipient.

        :return: Updates QuerySet as soft-deleted.
        """
        qs = self.active()
        if user:
            qs.filter(recipient=user)

        soft_delete = getattr(settings, 'NOTIFY_SOFT_DELETE', True)

        if soft_delete:
            qs.update(deleted=True)
        else:
            qs.delete()

    def active_all(self, user=None):
        """
        Method to soft-delete all notifications of a User (if supplied)

        :param user: Notification recipient.

        :return: Updates QuerySet as soft-deleted.
        """
        qs = self.deleted()
        if user:
            qs.filter(recipient=user)
        qs.update(deleted=False)

    def deleted(self):
        """
        QuerySet ``filter()`` for retrieving soft-deleted notifications.

        :return: Soft deleted notification filter()
        """
        return self.filter(deleted=True)


@python_2_unicode_compatible
class Notification(models.Model):

    """
    **Notification Model for storing notifications. (Yeah, too obvious)**

    This model is pretty-much a replica of ``django-notifications``'s
    model. The newly added fields just adds a feature to allow anonymous
    ``actors``, ``targets`` and ``object``.

    **Attributes**:
        :recipient:     The user who receives notification.
        :verb:          Action performed by actor (not necessarily).
        :description:   Option description for your notification.

        :actor_text:    Anonymous actor who is not in content-type.
        :actor_url:     Since the actor is not in content-type,
                        a custom URL for it.

        *...Same for target and obj*.

        :nf_type:   | Each notification is different, they must be formatted
                    | differently during HTML rendering. For this, each
                    | notification gets to carry it own *notification type*.
                    |
                    | This notification type will be used to search
                    | the special template for the notification located at
                    | ``notifications/includes/NF_TYPE.html`` of your
                    | template directory.
                    |
                    | The main reason to add this field is to save you
                    | from the pain of writing ``if...elif...else`` blocks
                    | in your template file just for handling how
                    | notifications will get rendered.
                    |
                    | With this, you can just save template for an individual
                    | notification type and call the *template-tag* to render
                    | all notifications for you without writing a single
                    | ``if...elif...else block``.
                    |
                    | You'll just need to do a
                    | ``{% render_notifications using NOTIFICATION_OBJ %}``
                    | and you'll get your notifications rendered.
                    |
                    | By default, every ``nf_type`` is set to ``default``.

    :extra:     **JSONField**, holds other optional data you want the
                notification to carry in JSON format.

    :deleted:   Useful when you want to *soft delete* your notifications.

    """

    recipient = models.ForeignKey(settings.AUTH_USER_MODEL,
                                  related_name='notifications',
                                  on_delete=models.CASCADE,
                                  verbose_name=_('Notification receiver'))

    # actor attributes.
    actor_content_type = models.ForeignKey(
        ContentType, null=True, blank=True,
        related_name='notify_actor', on_delete=models.CASCADE,
        verbose_name=_('Content type of actor object'))

    actor_object_id = models.PositiveIntegerField(
        null=True, blank=True,
        verbose_name=_('ID of the actor object'))

    actor_content_object = GenericForeignKey('actor_content_type',
                                             'actor_object_id')

    actor_text = models.CharField(
        max_length=200, blank=True, null=True,
        verbose_name=_('Anonymous text for actor'))

    actor_url_text = models.CharField(
        blank=True, null=True, max_length=200,
        verbose_name=_('Anonymous URL for actor'))

    # basic details.
    verb = models.CharField(max_length=200,
                            verbose_name=_('Verb of the action'))

    description = models.CharField(
        max_length=255, blank=True, null=True,
        verbose_name=_('Description of the notification'))

    nf_type = models.CharField(max_length=40, default='default',
                               verbose_name=_('Type of notification'))

    # TODO: Add a field to store notification cover images.

    # target attributes.
    target_content_type = models.ForeignKey(
        ContentType, null=True, blank=True,
        related_name='notify_target', on_delete=models.CASCADE,
        verbose_name=_('Content type of target object'))

    target_object_id = models.PositiveIntegerField(
        null=True, blank=True,
        verbose_name=_('ID of the target object'))

    target_content_object = GenericForeignKey('target_content_type',
                                              'target_object_id')

    target_text = models.CharField(
        max_length=200, blank=True, null=True,
        verbose_name=_('Anonymous text for target'))

    target_url_text = models.CharField(
        blank=True, null=True, max_length=200,
        verbose_name=_('Anonymous URL for target'))

    # obj attributes.
    obj_content_type = models.ForeignKey(
        ContentType, null=True, blank=True,
        related_name='notify_object', on_delete=models.CASCADE,
        verbose_name=_('Content type of action object'))

    obj_object_id = models.PositiveIntegerField(
        null=True, blank=True,
        verbose_name=_('ID of the target object'))

    obj_content_object = GenericForeignKey('obj_content_type', 'obj_object_id')

    obj_text = models.CharField(
        max_length=200, blank=True, null=True,
        verbose_name=_('Anonymous text for action object'))

    obj_url_text = models.CharField(
        blank=True, null=True, max_length=200,
        verbose_name=_('Anonymous URL for action object'))

    extra = JSONField(null=True, blank=True,
                      verbose_name=_('JSONField to store addtional data'))

    # Advanced details.
    created = models.DateTimeField(auto_now=False, auto_now_add=True)
    read = models.BooleanField(default=False,
                               verbose_name=_('Read status'))
    deleted = models.BooleanField(default=False,
                                  verbose_name=_('Soft delete status'))

    objects = NotificationQueryset.as_manager()

    class Meta(object):
        ordering = ('-created', )

    def __str__(self):
        ctx = {
            'actor': self.actor or self.actor_text,
            'verb': self.verb,
            'description': self.description,
            'target': self.target or self.target_text,
            'obj': self.obj or self.obj_text,
            'at': timesince(self.created),
        }

        if ctx['actor']:
            if not ctx['target']:
                return _("{actor} {verb} {at} ago").format(**ctx)
            elif not ctx['obj']:
                return _("{actor} {verb} on {target} {at} ago").format(**ctx)
            elif ctx['obj']:
                return _(
                    "{actor} {verb} {obj} on {target} {at} ago").format(**ctx)
        return _("{description} -- {at} ago").format(**ctx)

    def mark_as_read(self):
        """
        Marks notification as read
        """
        if self.read == True:
            return Response.HTTP_202_ACCEPTED
        self.read = True
        self.save()
        return Response.HTTP_200_OK

    def mark_as_unread(self):
        """
        Marks notification as unread.
        """
        if self.read == False:
            return Response.HTTP_202_ACCEPTED
        self.read = False
        self.save()
        return Response.HTTP_200_OK

    @cached_property
    def actor(self):
        """
        Property to return actor object/text to keep things DRY.

        :return: Actor object or Text or None.
        """
        return self.actor_content_object or self.actor_text

    @cached_property
    def actor_url(self):
        """
        Property to return permalink of the actor.
        Uses ``get_absolute_url()``.

        If ``get_absolute_url()`` method fails, it tries to grab URL
        from ``actor_url_text``, if it fails again, returns None (null).

        :return: URL for the actor.
        """
        try:
            url = self.actor_content_object.get_absolute_url()
        except AttributeError:
            url = self.actor_url_text or None
        return url

    @cached_property
    def target(self):
        """
        See ``actor`` property

        :return: Target object or Text or None
        """
        return self.target_content_object or self.target_text

    @cached_property
    def target_url(self):
        """
        See ``actor_url`` property.

        :return: URL for the target.
        """
        try:
            url = self.target_content_object.get_absolute_url()
        except AttributeError:
            url = self.target_url_text or None
        return url

    @cached_property
    def obj(self):
        """
        See ``actor`` property.

        :return: Action Object or Text or None.
        """
        return self.obj_content_object or self.obj_text

    @cached_property
    def obj_url(self):
        """
        See ``actor_url`` property.

        :return: URL for Action Object.
        """
        try:
            url = self.obj_content_object.get_absolute_url()
        except AttributeError:
            url = self.obj_url_text or None
        return url

    @staticmethod
    def do_escape(obj):
        """
        Method to HTML escape an object or set it to None conditionally.
        performs ``force_text()`` on the argument so that a foreignkey gets
        serialized? and spit out the ``__str__`` output instead of an Object.

        :param obj: Object to escape.

        :return: HTML escaped and JSON-friendly data.
        """
        return escape(force_text(obj)) if obj else None

    def as_json(self):
        """
        Notification data in a Python dictionary to which later gets
        supplied to JSONResponse so that it gets JSON serialized
        the *django-way*

        :return: Dictionary format of the QuerySet object.
        """
        data = {
            "id": self.id,
            "actor": self.do_escape(self.actor),
            "actor_url": self.do_escape(self.actor_url),
            "verb": self.do_escape(self.verb),
            "description": self.do_escape(self.description),
            "read": self.read,
            "nf_type": self.do_escape(self.nf_type),
            "target": self.do_escape(self.target),
            "target_url": self.do_escape(self.target_url),
            "obj": self.do_escape(self.obj),
            "obj_url": self.do_escape(self.obj_url),
            "created": self.created,
            "data": self.extra,
        }
        return data



from django.test import TestCase, override_settings
from django.contrib.auth import get_user_model
from notify.models import Notification
from notify.signals import notify
from django.utils import timezone
from django.core.urlresolvers import reverse
import json
from django.template import Template, Context, RequestContext
from django.test.client import RequestFactory
from django.contrib.auth.models import AnonymousUser

User = get_user_model()


class NotificationManagerTest(TestCase):

    def setUp(self):
        self.recipient = User.objects.create(username='recipient',
                                             email='recipient@test.com',
                                             password='pwd@recipient')
        self.actor = User.objects.create(username='actor',
                                         email='actor@test.com',
                                         password='pwd@actor')
        self.nf_count = 10

        self.assertEqual(self.recipient.id, 1)
        self.assertEqual(self.actor.id, 2)

        for x in range(self.nf_count):
            notify.send(User, recipient=self.recipient, actor=self.actor,
                        verb='followed you')

    def test_unread(self):
        self.assertEqual(Notification.objects.unread().count(), self.nf_count)
        nf = Notification.objects.filter(recipient=self.recipient).first()
        nf.mark_as_read()
        self.assertEqual(Notification.objects.unread().count(),
                         self.nf_count - 1)

        for n in Notification.objects.unread():
            self.assertFalse(n.read)

    def test_read(self):
        self.assertEqual(Notification.objects.unread().count(), self.nf_count)
        nf = Notification.objects.filter(recipient=self.recipient).first()
        nf.mark_as_read()
        self.assertEqual(Notification.objects.read().count(), 1)

        for n in Notification.objects.read():
            self.assertTrue(n.read)

    def test_read_all(self):
        self.assertEqual(Notification.objects.unread().count(), self.nf_count)
        recipient_nfs = Notification.objects.filter(recipient=self.recipient)
        recipient_nfs.read_all()

        self.assertEqual(Notification.objects.read().count(), self.nf_count)
        self.assertEqual(Notification.objects.unread().count(), 0)

    def test_unread_all(self):
        self.assertEqual(Notification.objects.unread().count(), self.nf_count)

        recipient_nfs = Notification.objects.filter(recipient=self.recipient)
        recipient_nfs.read_all()

        self.assertEqual(Notification.objects.read().count(), self.nf_count)
        self.assertEqual(Notification.objects.unread().count(), 0)

        recipient_nfs.unread_all()

        self.assertEqual(Notification.objects.read().count(), 0)
        self.assertEqual(Notification.objects.unread().count(), self.nf_count)

    def test_delete_all_soft(self):
        self.assertEqual(Notification.objects.unread().count(), self.nf_count)

        nf = Notification.objects.filter(recipient=self.recipient).first()
        # Mark single notification as read
        nf.mark_as_read()

        self.assertEqual(Notification.objects.read().count(), 1)
        self.assertEqual(Notification.objects.unread().count(),
                         self.nf_count - 1)

        self.assertEqual(Notification.objects.active().count(), self.nf_count)
        self.assertEqual(Notification.objects.deleted().count(), 0)

        Notification.objects.delete_all()
        # Mark all as deleted
        self.assertEqual(Notification.objects.read().count(), 0)
        self.assertEqual(Notification.objects.unread().count(), 0)
        self.assertEqual(Notification.objects.active().count(), 0)
        self.assertEqual(Notification.objects.deleted().count(), self.nf_count)

        Notification.objects.active_all()
        # Mark all as active
        self.assertEqual(Notification.objects.read().count(), 1)
        self.assertEqual(Notification.objects.unread().count(),
                         self.nf_count - 1)
        self.assertEqual(Notification.objects.active().count(), self.nf_count)
        self.assertEqual(Notification.objects.deleted().count(), 0)

    @override_settings(NOTIFY_SOFT_DELETE=False)
    def test_delete_all_hard(self):
        self.assertEqual(Notification.objects.active().count(), self.nf_count)

        # Hard delete all notifications.
        Notification.objects.delete_all()
        self.assertEqual(Notification.objects.active().count(), 0)
        self.assertEqual(Notification.objects.deleted().count(), 0)


class NotificationTest(TestCase):

    def setUp(self):
        self.no_of_users = 10

        users = []
        for i in range(self.no_of_users):
            users.append(
                User(username='user-%r' % i, password='pwd@user%r' % i,
                     email='user%r@test.com' % i))
        User.objects.bulk_create(users)

        self.actor = User.objects.create(username='actor',
                                         password='pwd@actor',
                                         email='actor@test.com')

        self.recipient = User.objects.get(username='user-0')
        self.recipient_list = User.objects.filter(
            username__startswith='u').order_by('id')

    def test_created_users(self):
        users = User.objects.all()
        self.assertEqual(users.count(), self.no_of_users + 1)
        self.assertEqual(self.recipient_list.count(), self.no_of_users)

    def test_single_user_notify(self):
        notify.send(User, recipient=self.recipient, actor=self.actor,
                    verb='poked you')
        notification = Notification.objects.get(pk=1)
        self.assertEqual(notification.recipient_id, self.recipient.id)
        timedelta = timezone.now() - notification.created
        self.assertLessEqual(timedelta.seconds, 60)

    def test_multiple_user_notify(self):
        notify.send(User, recipient_list=list(self.recipient_list),
                    actor=self.actor, verb='uploaded a new video')
        notifications = Notification.objects.filter(verb__startswith='u')
        self.assertEqual(notifications.count(), self.no_of_users)

        username_list = [u.username for u in self.recipient_list]

        for nf in notifications:
            self.assertIn(nf.recipient.username, username_list)
            timedelta = timezone.now() - nf.created
            self.assertLessEqual(timedelta.seconds, 60)


class NotificationViewTest(TestCase):

    def setUp(self):
        self.no_of_users = 10

        # Create recipients
        users = []
        for i in range(self.no_of_users):
            u = User(username='user-%r' % i, email='user%r@test.com' % i)
            u.set_password('pwd@user')
            users.append(u)
        User.objects.bulk_create(users)

        # Create actor
        actor = User.objects.create(username='actor',
                                    email='actor@test.com')
        actor.set_password('pwd@actor')
        actor.save()
        self.actor = User.objects.get(username="actor")

        self.recipient = User.objects.get(username='user-0')

        self.recipient_list = User.objects.filter(
            username__startswith='u').order_by('id')

        # Send notifications to all recipients.
        notify.send(User, recipient_list=list(self.recipient_list),
                    actor=self.actor, verb='wrote a new blog post.')

        # Send some more notifications to a specific user.
        self.recipient_nf_count = 11
        for i in range(10):
            notify.send(User, recipient=self.recipient, actor_text='You',
                        verb='reached level %r' % i)

        # Login this recipient
        self.assertTrue(self.client.login(username="user-0",
                                          password="pwd@user"))

    def test_created_users(self):
        users = User.objects.all()
        self.assertEqual(users.count(), self.no_of_users + 1)
        self.assertEqual(self.recipient_list.count(), self.no_of_users)
        self.assertEqual(self.recipient.id, 1)
        self.assertTrue(self.recipient.is_active)

    def test_notifications_view(self):
        response = self.client.get(reverse('notifications:all'))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context['notifications'].count(),
                         self.recipient.notifications.count())
        self.assertEqual(response.context['notifications'].count(),
                         self.recipient_nf_count)
        self.assertTemplateUsed(response, 'notifications/all.html')

    def test_mark_notification_view_non_ajax_without_next_parameter(self):
        self.assertEqual(self.recipient.notifications.unread().count(),
                         self.recipient_nf_count)

        first_nf = self.recipient.notifications.first()
        first_nf_id = first_nf.id
        self.assertFalse(first_nf.read)

        # Marking as read.
        response = self.client.post(reverse('notifications:mark'),
                                    {'action': 'read', 'id': first_nf_id})
        self.assertRedirects(response, reverse('notifications:all'))

        nf = Notification.objects.get(pk=first_nf_id)
        self.assertTrue(nf.read)

        self.assertEqual(self.recipient.notifications.unread().count(),
                         self.recipient_nf_count - 1)

        # Marking as unread
        response_2 = self.client.post(reverse('notifications:mark'),
                                      {'action': 'unread', 'id': first_nf_id})
        self.assertRedirects(response_2, reverse('notifications:all'))
        nf_2 = Notification.objects.get(pk=first_nf_id)
        self.assertFalse(nf_2.read)
        self.assertEqual(self.recipient.notifications.unread().count(),
                         self.recipient_nf_count)

    def test_mark_notification_view_non_ajax_with_next_parameter(self):
        self.assertEqual(self.recipient.notifications.unread().count(),
                         self.recipient_nf_count)

        first_nf = self.recipient.notifications.first()
        first_nf_id = first_nf.id
        self.assertFalse(first_nf.read)

        # Marking as read
        ctx = {'action': 'read', 'id': first_nf_id, 'next': '/notifications/'}
        response = self.client.post(reverse('notifications:mark'), ctx)
        self.assertRedirects(response, 'http://testserver/notifications/',
                             target_status_code=404, status_code=302)

        nf = Notification.objects.get(pk=first_nf_id)
        self.assertTrue(nf.read)

        self.assertEqual(self.recipient.notifications.unread().count(),
                         self.recipient_nf_count - 1)

        # Marking as unread
        ctx2 = {'action': 'unread', 'id': first_nf_id,
                'next': '/notifications/'}
        response_2 = self.client.post(reverse('notifications:mark'), ctx2)
        self.assertRedirects(response_2, 'http://testserver/notifications/',
                             target_status_code=404, status_code=302)
        nf_2 = Notification.objects.get(pk=first_nf_id)
        self.assertFalse(nf_2.read)
        self.assertEqual(self.recipient.notifications.unread().count(),
                         self.recipient_nf_count)

    def test_mark_notification_view_ajax(self):
        self.assertEqual(self.recipient.notifications.unread().count(),
                         self.recipient_nf_count)

        first_nf = self.recipient.notifications.first()
        first_nf_id = first_nf.id
        self.assertFalse(first_nf.read)

        # Marking as read
        ctx = {'action': 'read', 'id': first_nf_id}
        response = self.client.post(reverse('notifications:mark'), ctx,
                                    HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 200)

        nf = Notification.objects.get(pk=first_nf_id)
        self.assertTrue(nf.read)

        self.assertEqual(self.recipient.notifications.unread().count(),
                         self.recipient_nf_count - 1)

        # Marking as unread
        ctx = {'action': 'unread', 'id': first_nf_id}
        response_2 = self.client.post(reverse('notifications:mark'), ctx,
                                      HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response_2.status_code, 200)
        nf_2 = Notification.objects.get(pk=first_nf_id)
        self.assertFalse(nf_2.read)
        self.assertEqual(self.recipient.notifications.unread().count(),
                         self.recipient_nf_count)

    def test_mark_notification_view_for_arbitrary_user(self):
        """
        Our purpose here is to check whether user A can mark user B's
        notification as read or unread.
        """
        self.assertEqual(self.recipient.notifications.unread().count(),
                         self.recipient_nf_count)
        self.assertNotEqual(self.recipient.id, 2)

        nf_of_other_user = Notification.objects.get(recipient_id=2)
        self.assertFalse(nf_of_other_user.read)

        ctx = {'action': 'read', 'id': nf_of_other_user.id}

        response = self.client.post(reverse('notifications:mark'), ctx)

        self.assertEqual(response.status_code, 400)

        nf_of_other_user = Notification.objects.get(recipient_id=2)
        self.assertFalse(nf_of_other_user.read)
        self.assertEqual(self.recipient.notifications.unread().count(),
                         self.recipient_nf_count)

    # I'll escape the with and without `next` parameter test as it is
    # handled by the same functions, it is redundant to write them.

    def test_mark_all_notification_view_non_ajax(self):
        self.assertEqual(self.recipient.notifications.unread().count(),
                         self.recipient_nf_count)

        # Mark all as read
        ctx = {'action': 'read'}
        response = self.client.post(reverse('notifications:mark_all'), ctx)
        self.assertRedirects(response, reverse('notifications:all'))

        self.assertEqual(self.recipient.notifications.unread().count(), 0)
        self.assertEqual(self.recipient.notifications.read().count(),
                         self.recipient.notifications.count())

        # Mark all as unread
        ctx2 = {'action': 'unread'}
        response2 = self.client.post(reverse('notifications:mark_all'), ctx2)
        self.assertRedirects(response2, reverse('notifications:all'))

        self.assertEqual(self.recipient.notifications.read().count(), 0)
        self.assertEqual(self.recipient.notifications.unread().count(),
                         self.recipient.notifications.count())

    def test_mark_all_notification_view_ajax(self):
        self.assertEqual(self.recipient.notifications.unread().count(),
                         self.recipient_nf_count)

        # Mark all as read
        ctx = {'action': 'read'}
        response = self.client.post(reverse('notifications:mark_all'), ctx,
                                    HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 200)

        self.assertEqual(self.recipient.notifications.unread().count(), 0)
        self.assertEqual(self.recipient.notifications.read().count(),
                         self.recipient.notifications.count())

        # Mark all as unread
        ctx2 = {'action': 'unread'}
        response2 = self.client.post(reverse('notifications:mark_all'), ctx2,
                                     HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response2.status_code, 200)

        self.assertEqual(self.recipient.notifications.read().count(), 0)
        self.assertEqual(self.recipient.notifications.unread().count(),
                         self.recipient.notifications.count())

    def test_soft_delete_notification_view_non_ajax(self):
        self.assertEqual(self.recipient.notifications.active().count(),
                         self.recipient_nf_count)

        first_nf = self.recipient.notifications.first()
        first_nf_id = first_nf.id
        self.assertFalse(first_nf.deleted)

        # Soft delete notification
        ctx = {'id': first_nf_id}
        response = self.client.post(reverse('notifications:delete'), ctx)
        self.assertRedirects(response, reverse('notifications:all'))

        nf = Notification.objects.get(pk=first_nf_id)
        self.assertTrue(nf.deleted)

        self.assertEqual(self.recipient.notifications.active().count(),
                         self.recipient_nf_count - 1)

    def test_soft_delete_notification_view_ajax(self):
        self.assertEqual(self.recipient.notifications.active().count(),
                         self.recipient_nf_count)

        first_nf = self.recipient.notifications.first()
        first_nf_id = first_nf.id
        self.assertFalse(first_nf.deleted)

        # Soft delete notification
        ctx = {'id': first_nf_id}
        response = self.client.post(reverse('notifications:delete'), ctx,
                                    HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertEqual(response.status_code, 200)

        nf = Notification.objects.get(pk=first_nf_id)
        self.assertTrue(nf.deleted)

        self.assertEqual(self.recipient.notifications.active().count(),
                         self.recipient_nf_count - 1)

    @override_settings(NOTIFY_SOFT_DELETE=False)
    def test_hard_delete_notification_view(self):
        self.assertEqual(self.recipient.notifications.all().count(),
                         self.recipient_nf_count)

        first_nf = self.recipient.notifications.first()
        first_nf_id = first_nf.id

        # Hard delete notification
        ctx = {'id': first_nf_id}
        response = self.client.post(reverse('notifications:delete'), ctx)
        self.assertRedirects(response, reverse('notifications:all'))

        self.assertRaises(Notification.DoesNotExist,
                          Notification.objects.get,
                          pk=first_nf_id)

        self.assertEqual(self.recipient.notifications.all().count(),
                         self.recipient_nf_count - 1)

    def test_delete_notification_view_for_arbitrary_user(self):
        """
        Our purpose here is to check whether user A can mark user B's
        notification as read or unread.
        """
        self.assertNotEqual(self.recipient.id, 2)

        nf_of_other_user = Notification.objects.get(recipient_id=2)
        self.assertFalse(nf_of_other_user.deleted)

        ctx = {'id': nf_of_other_user.id}

        response = self.client.post(reverse('notifications:mark'), ctx)

        self.assertEqual(response.status_code, 400)

        nf_of_other_user = Notification.objects.get(recipient_id=2)
        self.assertFalse(nf_of_other_user.deleted)

        self.assertEqual(self.recipient.notifications.active().count(),
                         self.recipient_nf_count)

    def test_update_notifications_view(self):
        self.assertEqual(self.recipient.notifications.active().count(),
                         self.recipient_nf_count)
        nf = Notification.objects.get(pk=12)
        self.assertEqual(nf.recipient.username, 'user-0')

        url = "{}?flag=12".format(reverse('notifications:update'))

        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        resp = json.loads(response.content.decode('utf-8'))

        self.assertTrue(resp['success'])
        self.assertEqual(resp['unread_count'],
                         self.recipient.notifications.unread().count())
        notifications = self.recipient.notifications.active().filter(
            id__gt=12)
        self.assertEqual(resp['retrieved'],
                         notifications.count())

        for json_nf in resp['notifications']:
            self.assertIn(json_nf['id'],
                          [nf.id for nf in notifications])

    def test_update_view_with_wrong_flag_value(self):
        nf = Notification.objects.get(pk=3)
        self.assertNotEqual(nf.recipient.username, 'user-0')

        url = "{}?flag=this-wont-work".format(reverse('notifications:update'))
        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)
        resp = json.loads(response.content.decode('utf-8'))

        self.assertFalse(resp['success'])

        # A notification that does not exists
        # according to behavior described in #18, request should be successfu;
        # even if the notification ID doesn't exists/ belongs to the user.
        url2 = "{}?flag=123456789".format(reverse('notifications:update'))
        response2 = self.client.get(url2)

        self.assertEqual(response2.status_code, 200)
        resp2 = json.loads(response2.content.decode('utf-8'))

        # request is still successful
        self.assertTrue(resp2['success'])

    def test_read_and_redirect(self):
        """
        when users visits the view, they are redirected
        to the supplied next url. At the same time, the
        supplied notification id is marked as read.
        """
        nf = Notification.objects.get(pk=1)
        self.assertFalse(nf.read)
        self.assertEqual(nf.recipient.username, 'user-0')

        target_page = reverse('notifications:read_and_redirect', args=[nf.id])
        url = "{}?next=/home/".format(target_page)
        response = self.client.get(url)

        self.assertRedirects(response, "http://testserver/home/",
                             target_status_code=404, status_code=302)
        nf_new = Notification.objects.get(pk=1)
        self.assertTrue(nf_new.read)

    def test_read_and_redirect_unsafe_url(self):
        """
        when the ``next`` parameter is an unsafe url,
        user is redirected to the notifications page.
        """
        nf = Notification.objects.get(pk=1)
        self.assertFalse(nf.read)
        self.assertEqual(nf.recipient.username, 'user-0')

        target_page = reverse('notifications:read_and_redirect', args=[nf.id])
        url = "{}?next=file:///home/".format(target_page)
        response = self.client.get(url)

        self.assertRedirects(response, reverse('notifications:all'),
                             target_status_code=200, status_code=302)
        nf_new = Notification.objects.get(pk=1)
        self.assertTrue(nf_new.read)

    def test_read_and_redirect_arbitrary_notification_id(self):
        """
        When the users (attackers) change the notification id supplied
        in the URL manually to the notification id of another user (victim),
        the notification won't be marked as read.

        But the user (attacker) will still be redirected to the supplied
        next page if it is safe.
        """
        nf = Notification.objects.get(pk=3)
        self.assertFalse(nf.read)
        self.assertNotEqual(nf.recipient.username, 'user-0')

        target_page = reverse('notifications:read_and_redirect', args=[nf.id])
        url = "{}?next=/home/".format(target_page)
        response = self.client.get(url)

        self.assertRedirects(response, "http://testserver/home/",
                             target_status_code=404, status_code=302)

        nf_new = Notification.objects.get(pk=3)
        self.assertFalse(nf_new.read)


class NotificationTemplateTagTest(TestCase):

    RENDER_TEMPLATE = Template("""
    {% load notification_tags %}
    {% render_notifications using notifications for page %}
    """)

    RENDER_TEMPLATE_FOR_BOX = Template("""
    {% load notification_tags %}
    {% render_notifications using notifications for box %}
    """)

    JS_INCLUSION = Template("""
    {% load notification_tags  %}
    {% include_notify_js_variables %}
    """)

    USER_NOTIFICATIONS = Template("""
    {% load notification_tags %}
    {% user_notifications %}
    """)

    def setUp(self):
        user = User(username='user', email='user@test.com')
        user.set_password('pwd@user')
        user.save()

        self.user = User.objects.get(username='user')

        actor = User.objects.create_user('actor', 'actor@test.com',
                                         'pwd@actor')

        for x in range(10):
            notify.send(User, recipient=self.user, actor=actor,
                        verb='followed you', nf_type='followed_you')

        factory = RequestFactory()
        self.request = factory.get('/foobar/')

    def test_render_template_tag(self):
        notify.send(User, recipient=self.user,
                    actor_text='Joe', verb='followed you')
        nf_list = Notification.objects.filter(recipient=self.user).active()
        rendered = self.RENDER_TEMPLATE.render(
            Context({'notifications': nf_list}))
        self.assertIn('followed you', rendered)

    def test_render_template_tag_for_box(self):
        notify.send(User, recipient=self.user,
                    actor_text='Joe', verb='followed you')
        nf_list = Notification.objects.filter(recipient=self.user).active()
        rendered = self.RENDER_TEMPLATE_FOR_BOX.render(
            Context({'notifications': nf_list}))
        # to make things differentiable we'll use n-rTt-bx as a flag
        # in the default ``box`` template.
        self.assertIn('n-rTt-bx', rendered)
        self.assertIn('followed you', rendered)

    def test_render_template_tag_with_empty_notifications(self):
        nf_list = Notification.objects.filter(recipient=self.user).none()
        rendered = self.RENDER_TEMPLATE.render(
            Context({'notifications': nf_list}))
        self.assertIn('No notifications yet', rendered)

    def test_render_template_tag_without_context(self):
        rendered = self.RENDER_TEMPLATE.render(Context({}))
        self.assertIn('No notifications yet', rendered)

    def test_js_inclusion_tag(self):
        self.request.user = self.user
        rendered = self.JS_INCLUSION.render(RequestContext(self.request))
        self.assertIn('<script type="text/javascript">', rendered)

    def test_js_inclusion_tag_anonymous_user(self):
        self.request.user = AnonymousUser()
        rendered = self.JS_INCLUSION.render(RequestContext(self.request))
        self.assertNotIn('<script type="text/javascript">', rendered)

    def test_user_notifications(self):
        self.request.user = self.user
        rendered = self.USER_NOTIFICATIONS.render(RequestContext(self.request))
        self.assertIn('followed you', rendered)

    def test_user_notifications_for_anonymous_user(self):
        self.request.user = AnonymousUser()
        rendered = self.USER_NOTIFICATIONS.render(RequestContext(self.request))
        self.assertNotIn('followed you', rendered)

# TODO: Write test for test-worthy methods.



from django.conf import settings

# Class-selector of notification lists
NF_LIST_CLASS_SELECTOR = getattr(settings, 'NOTIFY_NF_LIST_CLASS_SELECTOR',
                                 '.notifications')

# Class-selector of mark read only notification lists
NF_READ_ONLY_LIST_CLASS_SELECTOR = getattr(settings, 'NOTIFY_NF_READ_ONLY_LIST_CLASS_SELECTOR', '.read-notifications')

# Class-selector of mark unread only notification lists
NF_UNREAD_ONLY_LIST_CLASS_SELECTOR = getattr(settings, 'NOTIFY_NF_READ_ONLY_LIST_CLASS_SELECTOR', '.unread-notifications')

# Class-selector for individual notifications.
SINGLE_NF_CLASS_SELECTOR = getattr(settings, 'NOTIFY_SINGLE_NF_CLASS_SELECTOR',
                                   '.notification')

# Class-selector for notification list on notification box.
NF_BOX_CLASS_SELECTOR = getattr(settings, 'NOTIFY_NF_BOX_CLASS_SELECTOR',
                                '.notification-box-list')

# Class-selector for individual notification in the notification box.
SINGLE_NF_BOX_CLASS_SELECTOR = getattr(settings,
                                       'NOTIFY_SINGLE_NF_BOX_CLASS_SELECTOR',
                                       '.notification-box')

# Class-selector for element performing `mark` as read/unread action.
MARK_NF_CLASS_SELECTOR = getattr(settings, 'NOTIFY_MARK_NF_CLASS_SELECTOR',
                                 '.mark-notification')

# Class-selector for link element to change opacity on `mark` as read/unread action.
MARK_NF_LINK_CLASS_SELECTOR = getattr(settings, 'NOTIFY_NF_LINK_CLASS_SELECTOR', '.mark-link-notification')

# Class-selector for element performing `mark_all` as read/unread action.
MARK_ALL_NF_CLASS_SELECTOR = getattr(settings,
                                     'NOTIFY_MARK_ALL_NF_CLASS_SELECTOR',
                                     '.mark-all-notifications')

# Class of notification element when it's status is read.
READ_NF_CLASS = getattr(settings, 'NOTIFY_READ_NOTIFICATION_CSS', 'read')

# Class of notification element when it's status is unread.
UNREAD_NF_CLASS = getattr(settings, 'NOTIFY_UNREAD_NOTIFICATION_CSS', 'unread')

# Class-selector for element performing `delete` action.
DELETE_NF_CLASS_SELECTOR = getattr(settings, 'NOTIFY_DELETE_NF_CLASS_SELECTOR',
                                   '.delete-notification')

# The tests fails if SOFT_DELETE setting is fetched from this file.
# Ref: https://code.djangoproject.com/ticket/22071
# TODO: Find a solution that passes the tests.
# # Whether do delete notifications softly or not.
# SOFT_DELETE = getattr(settings, 'NOTIFY_SOFT_DELETE', True)


# Time interval between ajax calls for notification update.
UPDATE_TIME_INTERVAL = getattr(settings, 'NOTIFY_UPDATE_TIME_INTERVAL', 5000)



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-03-25 19:20
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('notify', '0002_auto_20151221_0737'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notification',
            name='actor_url_text',
            field=models.CharField(blank=True, max_length=200, null=True, verbose_name='Anonymous URL for actor'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='obj_url_text',
            field=models.CharField(blank=True, max_length=200, null=True, verbose_name='Anonymous URL for action object'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='target_url_text',
            field=models.CharField(blank=True, max_length=200, null=True, verbose_name='Anonymous URL for target'),
        ),
    ]



from django.template.loader import render_to_string


def render_notification(notification, render_target='', **extra):
    template_name = notification.nf_type
    template_dir = 'notifications/includes/'

    nf_ctx = {'notification': notification}
    nf_ctx.update(extra)

    suffix = '_{}'.format(render_target) if render_target and render_target != 'page' else ''
    templates = [
        "{0}{1}{2}.html".format(template_dir, template_name, suffix),
        "{0}default{1}.html".format(template_dir, suffix),
        "{0}{1}.html".format(template_dir, template_name),
        "{0}default.html".format(template_dir),
    ]

    return render_to_string(templates, nf_ctx)


def prefetch_relations(weak_queryset):
    """
    FROM: https://djangosnippets.org/snippets/2492/

    Consider such a model class::

        class Action(models.Model):
            actor_content_type = models.ForeignKey(ContentType,related_name='actor')
            actor_object_id = models.PositiveIntegerField()
            actor = GenericForeignKey('actor_content_type','actor_object_id')

    And dataset::

        Action(actor=user1).save()
        Action(actor=user2).save()

    This will hit the user table once for each action::

        [a.actor for a in Action.objects.all()]

    Whereas this will hit the user table once::

        [a.actor for a in prefetch_relations(Action.objects.all())]

    Actually, the example above will hit the database N+1 times,  where N is
    the number of actions. But with prefetch_relations(), the database will be
    hit N+1 times where N is the number of distinct content types.

    Note that prefetch_relations() is recursive.

    Here an example, making a list with prefetch_relations(), and then without
    prefetch_relations(). See the number of database hits after each test.

        In [1]: from django import db; from prefetch_relations import prefetch_relations

        In [2]: db.reset_queries()

        In [3]: x = [(a.actor, a.action_object, a.target) for a in prefetch_relations(Action.objects.all().order_by('-pk'))]

        In [4]: print len(db.connection.queries)
        34

        In [5]: db.reset_queries()

        In [6]: print len(db.connection.queries)
        0

        In [7]: x = [(a.actor, a.action_object, a.target) for a in Action.objects.all().order_by('-pk')]

        In [8]: print len(db.connection.queries)
        396
    """
    from django.contrib.contenttypes.models import ContentType
    from django.contrib.contenttypes.fields import GenericForeignKey

    # reverse model's generic foreign keys into a dict:
    # { 'field_name': generic.GenericForeignKey instance, ... }
    gfks = {}
    for name, gfk in weak_queryset.model.__dict__.items():
        if not isinstance(gfk, GenericForeignKey):
            continue
        gfks[name] = gfk

    data = {}
    for weak_model in weak_queryset:
        for gfk_name, gfk_field in gfks.items():
            related_content_type_id = getattr(
                weak_model,
                gfk_field.model._meta.get_field(gfk_field.ct_field).get_attname())
            if not related_content_type_id:
                continue
            related_content_type = ContentType.objects.get_for_id(related_content_type_id)
            related_object_id = int(getattr(weak_model, gfk_field.fk_field))

            if related_content_type not in data.keys():
                data[related_content_type] = []
            data[related_content_type].append(related_object_id)

    for content_type, object_ids in data.items():
        model_class = content_type.model_class()
        models = prefetch_relations(model_class.objects.filter(pk__in=object_ids).select_related())
        for model in models:
            for weak_model in weak_queryset:
                for gfk_name, gfk_field in gfks.items():
                    related_content_type_id = getattr(
                        weak_model,
                        gfk_field.model._meta.get_field(gfk_field.ct_field).get_attname())
                    if not related_content_type_id:
                        continue
                    related_content_type = ContentType.objects.get_for_id(related_content_type_id)
                    related_object_id = int(getattr(weak_model, gfk_field.fk_field))

                    if related_object_id != model.pk:
                        continue
                    if related_content_type != content_type:
                        continue

                    setattr(weak_model, gfk_name, model)

    return weak_queryset



# -*- encoding: utf-8 -*-
import os
from setuptools import setup, find_packages
import notify as app


install_requires = open('requirements.txt').read().splitlines()


def read(fname):
    try:
        return open(os.path.join(os.path.dirname(__file__), fname)).read()
    except IOError:
        return ''

setup(
    name="django-notify-x2",
    version=app.__version__,
    description='Notification sytem for Django',
    long_description=read('README.rst'),
    license='The MIT License',
    platforms=['OS Independent'],
    keywords='django notifications, notify, facebook like notifications, github like notifications',
    author='Vikas Yadav, Ryan Castner',
    author_email='castner.rr@gmail.com',
    url="https://github.com/audiolion/django-notify-x2",
    packages=find_packages(),
    package_data={'notify': ['static/notify/*js',
                             'templates/*.html',
                             'templates/notifications/*.html',
                             'templates/notifications/includes/*.html',
                             'templates/notifications/includes/*.js']},
    include_package_data=True,
    install_requires=install_requires,

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Environment :: Web Environment',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Intended Audience :: Developers',
    ]
)



from django import template
from django.core.urlresolvers import reverse
from django.utils.translation import ugettext as _
from .. import notify_settings
from ..utils import render_notification

register = template.Library()


class RenderNotificationsNode(template.Node):

    """
    Template node to parse the token supplied and then generate html
    using the template of corresponding notification.
    """

    @classmethod
    def handle_token(cls, parser, token):
        tokens = token.split_contents()

        if tokens[1] != 'using':
            raise template.TemplateSyntaxError(
                _("The second argument in %r must be 'for'") % (tokens[0]))

        if len(tokens) == 3:
            return cls(parser.compile_filter(tokens[2]))
        elif len(tokens) == 5:
            if tokens[3] != 'for':
                raise template.TemplateSyntaxError(
                    _("The second argument in %r must be 'for'") % (tokens[0]))
            return cls(obj=parser.compile_filter(tokens[2]), target=tokens[4])
        else:
            raise template.TemplateSyntaxError(
                _("{tag} takes 2 or 3 arguments, {len} given.").format(
                    tag=tokens[0], len=len(tokens)))

    def __init__(self, obj, target=''):
        self.obj = obj
        self.target = target

    def generate_html(self, notifications):
        """
        Generates rendered HTML content using supplied notifications.
        :param notifications: Notification QuerySet Object
        :return: Rendered HTML.
        """
        html_chunks = []
        for nf in notifications:
            extra = nf.as_json() if self.target == 'box' else {}
            html = render_notification(nf, render_target=self.target, **extra)
            html_chunks.append(html)
        if not html_chunks:
            html_chunks.append(_("<b>No notifications yet.</b>"))
        html_string = '\n'.join(html_chunks)
        return html_string

    def render(self, context):
        """
        Render method of the template tag, returns generated html content to
        the parent template where it was called.
        :param context: Template context.
        :return: Generated HTML content using notification queryset object.
        """
        notifications = self.obj.resolve(context)
        return self.generate_html(notifications)


@register.tag
def render_notifications(parser, token):
    """
    Example::
        {% render_notifications using NOTIFICATION_QUERYSET_OBJECT %}
        {% render_notifications using NOTIFICATION_QUERYSET_OBJECT for \
        NOTIFICATION_RENDERING_TARGET %}

    :param parser: default arg
    :param token: default arg
    :return: Rendered HTML content for supplied notification QuerySet.
    """
    return RenderNotificationsNode.handle_token(parser, token)


JS_INCLUSION_TEMPATE = 'notifications/includes/js_variables.html'


@register.inclusion_tag(JS_INCLUSION_TEMPATE, takes_context=True)
def include_notify_js_variables(context):
    """
    Inclusion template tag to include all JS variables required by the
    notify.js file on the HTML page around <script> tags.

    Example::
        {% include_notify_js_variables %}

    :return: Prepares context for rendering in the inclusion file.
    """
    ctx = {
        'user': context['request'].user,
        'update_notification': reverse('notifications:update'),
        'mark_notification': reverse('notifications:mark'),
        'mark_read_notification': reverse('notifications:mark_read'),
        'mark_unread_notification': reverse('notifications:mark_unread'),
        'mark_all_notification': reverse('notifications:mark_all'),
        'delete_notification': reverse('notifications:delete'),

        'nf_list_class_selector': notify_settings.NF_LIST_CLASS_SELECTOR,
        'nf_read_only_list_class_selector': notify_settings.NF_READ_ONLY_LIST_CLASS_SELECTOR,
        'nf_unread_only_list_class_selector': notify_settings.NF_UNREAD_ONLY_LIST_CLASS_SELECTOR,
        'nf_class_selector': notify_settings.SINGLE_NF_CLASS_SELECTOR,
        'nf_box_list_class_selector': notify_settings.NF_BOX_CLASS_SELECTOR,
        'nf_box_class_selector': notify_settings.SINGLE_NF_BOX_CLASS_SELECTOR,

        'nf_mark_selector': notify_settings.MARK_NF_CLASS_SELECTOR,
        'nf_mark_all_selector': notify_settings.MARK_ALL_NF_CLASS_SELECTOR,
        'nf_delete_selector': notify_settings.DELETE_NF_CLASS_SELECTOR,
        'nf_mark_link_selector': notify_settings.MARK_NF_LINK_CLASS_SELECTOR,

        'nf_read_class': notify_settings.READ_NF_CLASS,
        'nf_unread_class': notify_settings.UNREAD_NF_CLASS,

        'nf_update_time_interval': notify_settings.UPDATE_TIME_INTERVAL,
    }
    return ctx


class UserNotification(RenderNotificationsNode):

    """
    Returns notifications for a supplied user, can be used as shortcut
    for render_notifications
    """
    @classmethod
    def handle_token(cls, parser, token):
        tokens = token.split_contents()
        if len(tokens) == 1:
            # Uses the object parameter as flag
            return cls(obj='user-notification', target='page')

        if len(tokens) > 3:
            raise template.TemplateSyntaxError(_("Max arguments are two"))
        elif tokens[1] != 'for':
            raise template.TemplateSyntaxError(_("First argument must be 'for'"))
        elif not tokens[2]:
            raise template.TemplateSyntaxError(
                _("Second argument must either 'box' or 'page'"))
        else:
            return cls(obj='user-notification', target=tokens[2])

    def render(self, context):
        if self.obj == 'user-notification':
            request = context['request']
            user = request.user
            if user.is_authenticated():
                notifications = user.notifications.active().prefetch()
                return self.generate_html(notifications)
        return ''


@register.tag
def user_notifications(parser, token):
    """
    Example::
        {% user_notifications for NOTIFICATION_RENDERING_TARGET %}

    :param parser: default arg
    :param token: default arg
    :return: Rendered HTML content for supplied notification QuerySet.
    """
    return UserNotification.handle_token(parser, token)



from django import dispatch
from django.dispatch import receiver
from notify.models import Notification
from django.utils.translation import ugettext as _


notify = dispatch.Signal(providing_args=[
    'recipient', 'recipient_list',
    'actor', 'actor_text', 'actor_url',
    'verb', 'description', 'nf_type',
    'target', 'target_text', 'target_url',
    'obj', 'obj_text', 'obj_url',
    'extra',
])


def truncate(string, length):
    if string is not None and isinstance(string, str):
        if len(string) > length:
            return string[:length]
        return string


@receiver(notify, dispatch_uid='notify_user')
def notifier(sender, **kwargs):
    recipient = kwargs.pop('recipient', None)

    recipient_list = kwargs.pop('recipient_list', None)

    verb = kwargs.pop('verb', None)
    description = kwargs.pop('description', None)
    nf_type = kwargs.pop('nf_type', 'default')

    actor = kwargs.pop('actor', None)
    actor_text = kwargs.pop('actor_text', None)
    actor_url = kwargs.pop('actor_url', None)

    target = kwargs.pop('target', None)
    target_text = kwargs.pop('target_text', None)
    target_url = kwargs.pop('target_url', None)

    obj = kwargs.pop('obj', None)
    obj_text = kwargs.pop('obj_text', None)
    obj_url = kwargs.pop('obj_url', None)

    extra = kwargs.pop('extra', None)

    actor_text = truncate(actor_text, Notification._meta.get_field('actor_text').max_length)
    actor_url = truncate(actor_text, Notification._meta.get_field('actor_url_text').max_length)

    target_text = truncate(target_text, Notification._meta.get_field('target_text').max_length)
    target_url = truncate(target_text, Notification._meta.get_field('target_url_text').max_length)

    obj_text = truncate(obj_text, Notification._meta.get_field('obj_text').max_length)
    obj_url = truncate(obj_text, Notification._meta.get_field('obj_url_text').max_length)

    if recipient and recipient_list:
        raise TypeError(_("You must specify either a single recipient or a list"
                        " of recipients, not both."))
    elif not recipient and not recipient_list:
        raise TypeError(_("You must specify the recipient of the notification."))

    if not actor and not actor_text:
        raise TypeError(_("Actor not specified."))

    if not verb:
        raise TypeError(_("Verb not specified."))

    if recipient_list and not isinstance(recipient_list, list) and not isinstance(recipient_list, set):
        raise TypeError(_("Supplied recipient is not an instance of list or set."))

    if recipient:
        notification = Notification(
            recipient=recipient,
            verb=verb, description=description, nf_type=nf_type,
            actor_content_object=actor, actor_text=actor_text,
            actor_url_text=actor_url,

            target_content_object=target, target_text=target_text,
            target_url_text=target_url,

            obj_content_object=obj, obj_text=obj_text, obj_url_text=obj_url,
            extra=extra
        )
        saved_notification = notification.save()
    else:
        notifications = []
        for recipient in recipient_list:
            notifications.append(Notification(
                recipient=recipient,
                verb=verb, description=description, nf_type=nf_type,
                actor_content_object=actor, actor_text=actor_text,
                actor_url_text=actor_url,

                target_content_object=target, target_text=target_text,
                target_url_text=target_url,

                obj_content_object=obj, obj_text=obj_text,
                obj_url_text=obj_url,

                extra=extra
            ))
        saved_notification = Notification.objects.bulk_create(notifications)
    return saved_notification
